package at.bestsolution.rsd.rt.client.ws;

import java.io.IOException;
import java.util.Collections;
import java.util.HashMap;
import java.util.Map;
import java.util.function.BiConsumer;
import java.util.function.Function;

import javax.inject.Inject;
import javax.websocket.OnMessage;
import javax.websocket.OnOpen;
import javax.websocket.Session;

import org.eclipse.fx.core.log.FluentLogger;
import org.eclipse.fx.core.log.LoggerCreator;

import at.bestsolution.rsd.rt.client.InvokationUtil;
import at.bestsolution.rsd.rt.rpc.RPCMessageBuilder;
import at.bestsolution.rsd.rt.rpc.RPCMessageParser;
import at.bestsolution.rsd.rt.rpc.RPCRequestMessage;
import at.bestsolution.rsd.rt.rpc.RPCResponseMessage;

public class BaseWSClientEndpoint {
	private RPCMessageBuilder messageBuilder;
	private RPCMessageParser messageParser;
	private FluentLogger logger;
	private Session session;
	
	private final Map<String, BiConsumer<String, String>> responseHandlers = Collections.synchronizedMap(new HashMap<>());
	protected static final long DEFAULT_TIMEOUT = 10_000;
	private static final Function<String, Void> VOID_CONVERTER = (s) -> null;
	
	@Inject
	public void setRPCMessageBuilder(RPCMessageBuilder messageBuilder) {
		this.messageBuilder = messageBuilder;
	}

	@Inject
	public void setRPCMessageParser(RPCMessageParser messageParser) {
		this.messageParser = messageParser;
	}

	protected FluentLogger getLogger() {
		if (logger == null) {
			logger = FluentLogger.of(LoggerCreator.createLogger(getClass()));
		}
		return logger;
	}
	
	private void sendText(String text) {
		try {
			session.getBasicRemote().sendText(text);	
		} catch( IOException e ) {
			throw new RuntimeException(e);
		}
	}
	
	private RPCRequestMessage createRequestMessage(String methodId, Map<String, String> parameters) {
		return messageBuilder.request(methodId, parameters);
	}
	
	private void registerResponseHandler(String id, BiConsumer<String, String> responsHandler) {
		this.responseHandlers.put(id, responsHandler);
	}
	
	private void unregisterResponseHandler(String id) {
		this.responseHandlers.remove(id);
	}
	
	protected <O> O sendMessage(long timeout, String methodId, Map<String, String> parameters, Function<String, O> responseConverter) {
		return InvokationUtil.invoke(createRequestMessage(methodId, parameters), this::sendText, timeout, responseConverter, this::registerResponseHandler, this::unregisterResponseHandler);
	}

	protected void sendVoidMessage(long timeout, String methodId, Map<String, String> parameters) {
		InvokationUtil.invoke(createRequestMessage(methodId, parameters), this::sendText, timeout, VOID_CONVERTER, this::registerResponseHandler, this::unregisterResponseHandler);
	}
	
	@OnOpen
	public final void onWebSocketConnect(Session session) {
		this.session = session;
	}

	@OnMessage
	public final void receivedMessage(String message) {
		RPCResponseMessage response = messageParser.response(message);
		BiConsumer<String, String> consumer = responseHandlers.get(response.requestId());
		if (consumer == null) {
			getLogger().atWarning().log("Could not find receiver for '%s'. Maybe it already timed out.",
					response.requestId());
		}
		consumer.accept(new String(response.data()), null);
	}
}
